@@ -10,8 +10,9 @@ https://gitlab.com/keatontaylor/alexapy
 import asyncio
 import json
 import logging
-from typing import Any, Dict, Optional, Text
-from typing import List  # noqa pylint: disable=unused-import
+import math
+import time
+from typing import Any, Dict, List, Optional, Text
 
 import backoff
 from yarl import URL
@@ -66,6 +67,38 @@ class AlexaAPI:
                 ex,
             )
 
+    def update_login(self, login: AlexaLogin) -> bool:
+        """Update Login if it has changed.
+
+        Args
+            login (AlexaLogin): AlexaLogin to check
+
+        Returns
+            bool: True if change detected
+
+        """
+        if login != self._login or login.session != self._session:
+            _LOGGER.debug("New Login %s detected; replacing %s", login, self._login)
+            self._login = login
+            self._session = login.session
+            self._url: Text = "https://alexa." + login.url
+            self._login._headers["Referer"] = "{}/spa/index.html".format(self._url)
+            try:
+                assert self._login._cookies is not None
+                csrf = self._login._cookies["csrf"]
+                self._login._headers["csrf"] = csrf
+            except KeyError as ex:
+                if login.status.get("login_successful"):
+                    _LOGGER.warning(
+                        (
+                            "AlexaLogin session is missing required token: %s "
+                            "This may result in authorization errors, please report"
+                        ),
+                        ex,
+                    )
+            return True
+        return False
+
     @backoff.on_exception(
         backoff.expo,
         (AlexapyTooManyRequestsError, AlexapyConnectionError, ClientConnectionError),
@@ -80,6 +113,17 @@ class AlexaAPI:
         data: Optional[Dict[Text, Text]] = None,
         query: Optional[Dict[Text, Text]] = None,
     ) -> ClientResponse:
+        async with self._login._oauth_lock:
+            if self._login.expires_in and (self._login.expires_in - time.time() < 0):
+                _LOGGER.debug("Detected access token expiration; refreshing")
+                await self._login.refresh_access_token()
+                await self._login.exchange_token_for_cookies()
+                await self._login.save_cookiefile()
+        if method == "get":
+            if query and not query.get("_"):
+                query["_"] = math.floor(time.time() * 1000)
+            elif query is None:
+                query = {"_": math.floor(time.time() * 1000)}
         url: URL = URL(self._url + uri).update_query(query)
         # _LOGGER.debug("Trying %s: %s : with uri: %s data %s query %s",
         #               method,
@@ -96,6 +140,15 @@ class AlexaAPI:
                 query,
             )
             raise AlexapyLoginCloseRequested()
+        if not self._login.status.get("login_successful"):
+            _LOGGER.debug(
+                "Login error detected; ignoring %s request to %s with %s %s",
+                method,
+                uri,
+                data,
+                query,
+            )
+            raise AlexapyLoginError("Login error detected; not contacting API")
         if self._session.closed:
             raise AlexapyLoginError("Session is closed")
         response = await getattr(self._session, method)(
@@ -113,7 +166,9 @@ class AlexaAPI:
             response.reason,
             response.content_type,
         )
+        self._login.stats["api_calls"] += 1
         if response.status == 401:
+            self._login.status["login_successful"] = False
             raise AlexapyLoginError(response.reason)
         if response.status == 429:
             raise AlexapyTooManyRequestsError(response.reason)
@@ -154,6 +209,12 @@ class AlexaAPI:
         data: Optional[Dict[Text, Text]] = None,
         query: Optional[Dict[Text, Text]] = None,
     ) -> ClientResponse:
+        async with login._oauth_lock:
+            if login.expires_in and (login.expires_in - time.time() < 0):
+                _LOGGER.debug("Detected access token expiration; refreshing")
+                await login.refresh_access_token()
+                await login.exchange_token_for_cookies()
+                await login.save_cookiefile()
         session = login.session
         url: URL = URL("https://alexa." + login.url + uri).update_query(query)
         # _LOGGER.debug("Trying static %s: %s : with uri: %s data %s query %s",
@@ -171,6 +232,15 @@ class AlexaAPI:
                 query,
             )
             raise AlexapyLoginCloseRequested()
+        if not login.status.get("login_successful"):
+            _LOGGER.debug(
+                "Login error detected; ignoring %s request to %s with %s %s",
+                method,
+                uri,
+                data,
+                query,
+            )
+            raise AlexapyLoginError("Login error detected; not contacting API")
         if session.closed:
             raise AlexapyLoginError("Session is closed")
         response = await getattr(session, method)(
@@ -189,6 +259,7 @@ class AlexaAPI:
             response.content_type,
         )
         if response.status == 401:
+            login.status["login_successful"] = False
             raise AlexapyLoginError(response.reason)
         if response.status == 429:
             raise AlexapyTooManyRequestsError(response.reason)
@@ -212,6 +283,8 @@ class AlexaAPI:
             "@type": "com.amazon.alexa.behaviors.model.Sequence",
             "startNode": node_data,
         }
+        if queue_delay is None:
+            queue_delay = 1.5
         if queue_delay > 0:
             sequence_json["startNode"] = {
                 "@type": "com.amazon.alexa.behaviors.model.SerialNode",
@@ -275,6 +348,7 @@ class AlexaAPI:
         sequence: Text,
         customer_id: Optional[Text] = None,
         queue_delay: float = 1.5,
+        extra: Optional[Dict[Any, Any]] = None,
         **kwargs,
     ) -> None:
         """Send sequence command.
@@ -292,6 +366,7 @@ class AlexaAPI:
                                     for commands to queue together.
                                     Defaults to 1.5.
                                     Must be positive.
+        extra (Dict): Extra dictionary array; functionality undetermined
         **kwargs : Each named variable must match a recognized Amazon variable
                    within the operationPayload. Please see examples in
                    play_music, send_announcement, and send_tts.
@@ -316,6 +391,7 @@ class AlexaAPI:
         https://github.com/custom-components/alexa_media_player/wiki#sequence-commands-versions--100
 
         """
+        extra = extra or {}
         operation_payload = {
             "deviceType": self._device._device_type,
             "deviceSerialNumber": self._device.unique_id,
@@ -388,6 +464,39 @@ class AlexaAPI:
         }
         await self.run_behavior(node_data, queue_delay=queue_delay)
 
+    @_catch_all_exceptions
+    async def run_custom(
+        self,
+        text: Text,
+        customer_id: Optional[Text] = None,
+        queue_delay: float = 0,
+        extra: Optional[Dict[Any, Any]] = None,
+    ) -> None:
+        """Run Alexa skill.
+
+        This allows running exactly what you can say to alexa.
+
+        Args:
+            text (string): The full text you want alexa to execute.
+            customer_id (string): CustomerId to use for authorization. When none
+                             specified this defaults to the logged in user. Used
+                             with households where others may have their own
+                             music.
+            queue_delay (float, optional): The number of seconds to wait
+                                        for commands to queue together.
+                                        Defaults to 1.5.
+                                        Must be positive.
+            extra (Dict): Extra dictionary array; functionality undetermined
+
+        """
+        extra = extra or {}
+        await self.send_sequence(
+            "Alexa.TextCommand",
+            skillId="amzn1.ask.1p.tellalexa",
+            text=text,
+            queue_delay=queue_delay,
+        )
+
     @_catch_all_exceptions
     async def run_routine(
         self,
@@ -444,7 +553,10 @@ class AlexaAPI:
             if "utterance" not in automation["triggers"][0]["payload"]:
                 continue
             a_utterance = automation["triggers"][0]["payload"]["utterance"]
-            if a_utterance is not None and a_utterance.lower() == utterance.lower():
+            a_name = automation.get("name")
+            if (
+                a_utterance is not None and a_utterance.lower() == utterance.lower()
+            ) or (a_name and a_name.lower() == utterance.lower()):
                 automation_id = automation["automationId"]
                 sequence = automation["sequence"]
         if automation_id is None or sequence is None:
@@ -483,6 +595,7 @@ class AlexaAPI:
         customer_id: Optional[Text] = None,
         timer: Optional[int] = None,
         queue_delay: float = 1.5,
+        extra: Optional[Dict[Any, Any]] = None,
     ) -> None:
         """Play music based on search.
 
@@ -494,9 +607,13 @@ class AlexaAPI:
                              with households where others may have their own
                              music.
             timer (Optional[int]): Number of seconds to play before stopping.
-            queue_delay (float, optional): [description]. Defaults to 1.5.
+            queue_delay (float, optional): The number of seconds to wait
+                                   for commands to queue together.
+                                   Must be positive. Defaults to 1.5.
+            extra (Dict): Extra dictionary array; functionality undetermined
 
         """
+        extra = extra or {}
         customer_id = self._login.customer_id if customer_id is None else customer_id
         if timer:
             await self.send_sequence(
@@ -524,8 +641,10 @@ class AlexaAPI:
         sound_string_id: Text,
         customer_id: Optional[Text] = None,
         queue_delay: float = 1.5,
+        extra: Optional[Dict[Any, Any]] = None,
     ) -> None:
         """Play Alexa sound."""
+        extra = extra or {}
         await self.send_sequence(
             "Alexa.Sound",
             customer_id=self._login.customer_id if customer_id is None else customer_id,
@@ -692,6 +811,7 @@ class AlexaAPI:
         customer_id: Optional[Text] = None,
         targets: Optional[List[Text]] = None,
         queue_delay: float = 1.5,
+        extra: Optional[Dict[Any, Any]] = None,
     ) -> None:
         # pylint: disable=too-many-arguments
         """Send announcment to Alexa devices.
@@ -715,8 +835,10 @@ class AlexaAPI:
                                         for commands to queue together.
                                         Defaults to 1.5.
                                         Must be positive.
+        extra (Dict): Extra dictionary array; functionality undetermined
 
         """
+        extra = extra or {}
         display = (
             {"title": "", "body": ""}
             if method.lower() == "speak"
@@ -757,6 +879,7 @@ class AlexaAPI:
         title: Text = "AlexaAPI Message",
         customer_id: Optional[Text] = None,
         queue_delay: float = 1.5,
+        extra: Optional[Dict[Any, Any]] = None,
     ) -> None:
         """Send mobile push to Alexa app.
 
@@ -772,8 +895,10 @@ class AlexaAPI:
                                         for commands to queue together.
                                         Defaults to 1.5.
                                         Must be positive.
+        extra (Dict): Extra dictionary array; functionality undetermined
 
         """
+        extra = extra or {}
         await self.send_sequence(
             "Alexa.Notifications.SendMobilePush",
             customer_id=(
@@ -793,6 +918,7 @@ class AlexaAPI:
         title: Text = "AlexaAPI Dropin Notification",
         customer_id: Optional[Text] = None,
         queue_delay: float = 1.5,
+        extra: Optional[Dict[Any, Any]] = None,
     ) -> None:
         """Send dropin notification to Alexa app for Alexa device.
 
@@ -808,8 +934,10 @@ class AlexaAPI:
                                         for commands to queue together.
                                         Defaults to 1.5.
                                         Must be positive.
+        extra (Dict): Extra dictionary array; functionality undetermined
 
         """
+        extra = extra or {}
         await self.send_sequence(
             "Alexa.Notifications.DropIn",
             customer_id=(
@@ -1296,3 +1424,28 @@ class AlexaAPI:
         success = response.status == 200
         _LOGGER.debug("Success: %s Response: %s", success, response_json)
         return success
+
+    @staticmethod
+    @_catch_all_exceptions
+    async def force_logout() -> None:
+        """Force logout.
+
+        Raises
+            AlexapyLoginError: Raise AlexapyLoginError
+
+        """
+        raise AlexapyLoginError("Forced Logout")
+
+    @staticmethod
+    @_catch_all_exceptions
+    async def ping(login: AlexaLogin) -> Optional[Dict[Text, Any]]:
+        """Ping.
+
+        Args:
+        login (AlexaLogin): Successfully logged in AlexaLogin
+
+        Returns json
+
+        """
+        response = await AlexaAPI._static_request("get", login, "/api/ping",)
+        return await response if response else None
