@@ -8,25 +8,33 @@ For more details about this api, please refer to the documentation at
 https://gitlab.com/keatontaylor/alexapy
 """
 
-from json import JSONDecodeError
+import asyncio
+from http.cookies import SimpleCookie
+from json import JSONDecodeError, dumps
+import base64
+import secrets
 import logging
+import datetime
 import os
 import pickle
 import re
-from typing import Callable, List, Optional, Text, Tuple, Union
-from typing import Dict  # noqa pylint: disable=unused-import
+from binascii import Error
+from typing import Any, Callable, Dict, List, Optional, Text, Tuple, Union
 from urllib.parse import urlencode, urlparse
 
 import aiofiles
 from aiofiles import os as aioos
 from bs4 import BeautifulSoup
+import pyotp
 from simplejson import JSONDecodeError as SimpleJSONDecodeError
+from yarl import URL
 
 from alexapy import aiohttp
 from alexapy.aiohttp.client_exceptions import ContentTypeError
 
-from .const import EXCEPTION_TEMPLATE
-from .helpers import _catch_all_exceptions, delete_cookie, obfuscate
+from .const import EXCEPTION_TEMPLATE, APP_NAME, USER_AGENT
+from .errors import AlexapyPyotpInvalidKey
+from .helpers import _catch_all_exceptions, delete_cookie, hide_serial, obfuscate
 
 _LOGGER = logging.getLogger(__name__)
 
@@ -41,6 +49,7 @@ class AlexaLogin:
     password (string): Password for Amazon login account
     outputpath (function): Local path with write access for storing files
     debug (boolean): Enable additional debugging including debug file creation
+    otp_secret (string): TOTP Secret key for automatic 2FA filling
 
     """
 
@@ -51,12 +60,16 @@ class AlexaLogin:
         password: Text,
         outputpath: Callable[[Text], Text],
         debug: bool = False,
+        otp_secret: Text = "",
+        oauth: Optional[Dict[Any, Any]] = None,
     ) -> None:
         # pylint: disable=too-many-arguments,import-outside-toplevel
         """Set up initial connection and log in."""
         import ssl
+
         import certifi
 
+        oauth = oauth or {}
         prefix: Text = "alexa_media"
         self._prefix = "https://alexa."
         self._url: Text = url
@@ -70,9 +83,14 @@ class AlexaLogin:
         self._headers: Dict[Text, Text] = {}
         self._data: Optional[Dict[Text, Text]] = None
         self.status: Optional[Dict[Text, Union[Text, bool]]] = {}
+        self.stats: Optional[Dict[Text, Union[Text, bool]]] = {
+            "login_timestamp": datetime.datetime(1, 1, 1),
+            "api_calls": 0,
+        }
         self._cookiefile: List[Text] = [
             outputpath(".storage/{}.{}.pickle".format(prefix, email)),
             outputpath("{}.{}.pickle".format(prefix, email)),
+            outputpath(".storage/{}.{}.txt".format(prefix, email)),
         ]
         self._debugpost: Text = outputpath("{}{}post.html".format(prefix, email))
         self._debugget: Text = outputpath("{}{}get.html".format(prefix, email))
@@ -84,6 +102,12 @@ class AlexaLogin:
         self._create_session()
         self._close_requested = False
         self._customer_id: Optional[Text] = None
+        self._totp: Optional[pyotp.TOTP] = None
+        self.set_totp(otp_secret.replace(" ", ""))
+        self.access_token: Optional[Text] = oauth.get("access_token")
+        self.refresh_token: Optional[Text] = oauth.get("refresh_token")
+        self.expires_in: Optional[float] = oauth.get("expires_in")
+        self._oauth_lock: asyncio.Lock = asyncio.Lock()
 
     @property
     def email(self) -> Text:
@@ -123,37 +147,113 @@ class AlexaLogin:
             result += f"link{key}:{value[0]}\n"
         return result
 
-    async def login_with_cookie(self) -> None:
-        # pylint: disable=import-outside-toplevel
-        """Attempt to login after loading cookie."""
-        from requests.cookies import RequestsCookieJar
-        from collections import defaultdict
+    def set_totp(self, otp_secret: Text) -> Optional[pyotp.TOTP]:
+        """Enable a TOTP generator for the login.
 
-        cookies: Optional[RequestsCookieJar] = None
+        Args
+            otp_secret (Text): Secret. If blank, it will remove the TOTP entry.
+
+        Returns
+            Optional[pyotp.TOTP]: The pyotp TOTP object
+
+        """
+        if otp_secret:
+            _LOGGER.debug("Creating TOTP for %s", hide_serial(otp_secret))
+            try:
+                self._totp = pyotp.TOTP(otp_secret)
+                self.get_totp_token()
+            except Error as ex:
+                self._totp = None
+                _LOGGER.warning(
+                    "Error creating TOTP; %s likely invalid", hide_serial(otp_secret)
+                )
+                raise AlexapyPyotpInvalidKey(ex) from ex
+        else:
+            self._totp = None
+        return self._totp
+
+    def get_totp_token(self) -> Text:
+        """Generate Timed based OTP token.
+
+        Returns
+            Text: OTP for current time.
+
+        """
+        if self._totp:
+            token: Text = self._totp.now()
+            _LOGGER.debug("Generating OTP %s", token)
+            return token
+        _LOGGER.debug("Unable to generate OTP; 2FA app key not configured")
+        return ""
+
+    async def load_cookie(self, cookies_txt: Text = "") -> Optional[Dict[Text, Text]]:
+        # pylint: disable=import-outside-toplevel
+        """Load cookie from disk."""
+        from collections import defaultdict
+        import http.cookiejar
+
+        from requests.cookies import RequestsCookieJar
+
+        cookies: Optional[
+            Union[RequestsCookieJar, http.cookiejar.MozillaCookieJar]
+        ] = None
         numcookies: int = 0
         loaded: bool = False
         if self._cookiefile:
+            if cookies_txt:
+                _LOGGER.debug(
+                    "Saving passed in cookie to %s\n%s",
+                    self._cookiefile[0],
+                    repr(cookies_txt),
+                )
+                async with aiofiles.open(self._cookiefile[0], mode="w") as localfile:
+                    try:
+                        await localfile.write(cookies_txt)
+                    except (OSError, EOFError, TypeError, AttributeError) as ex:
+                        _LOGGER.debug(
+                            "Error saving passed in cookie to %s: %s",
+                            self._cookiefile[0],
+                            EXCEPTION_TEMPLATE.format(type(ex).__name__, ex.args),
+                        )
             for cookiefile in self._cookiefile:
+                _LOGGER.debug("Searching for cookies from %s", cookiefile)
+                if loaded:
+                    break
                 numcookies = 0
                 if not os.path.exists(cookiefile):
                     continue
                 if loaded and cookiefile != self._cookiefile[0]:
                     await delete_cookie(cookiefile)
-                _LOGGER.debug("Trying to load pickled cookie from file %s", cookiefile)
+                _LOGGER.debug("Trying to load cookie from file %s", cookiefile)
                 try:
                     async with aiofiles.open(cookiefile, "rb") as myfile:
                         cookies = pickle.loads(await myfile.read())
                         if self._debug:
                             _LOGGER.debug(
-                                "cookie loaded: %s %s", type(cookies), cookies
+                                "Pickled cookie loaded: %s %s", type(cookies), cookies
                             )
-                except (OSError, EOFError, pickle.UnpicklingError) as ex:
+                except (pickle.UnpicklingError):
+                    try:
+                        cookies = http.cookiejar.MozillaCookieJar(cookiefile)
+                        cookies.load(ignore_discard=True, ignore_expires=True)
+                        if self._debug:
+                            _LOGGER.debug(
+                                "Mozilla cookie loaded: %s %s", type(cookies), cookies
+                            )
+                    except (ValueError, http.cookiejar.LoadError) as ex:
+                        _LOGGER.debug(
+                            "Cookie %s is truncated: %s",
+                            cookiefile,
+                            EXCEPTION_TEMPLATE.format(type(ex).__name__, ex.args),
+                        )
+                        continue
+                except (OSError, EOFError) as ex:
                     _LOGGER.debug(
-                        "Error loading pickled cookie from %s: %s",
+                        "Error loading cookie from %s: %s",
                         cookiefile,
                         EXCEPTION_TEMPLATE.format(type(ex).__name__, ex.args),
                     )
-                # escape extra quote marks from Requests cookie
+                    continue
                 if isinstance(cookies, RequestsCookieJar):
                     _LOGGER.debug("Loading RequestsCookieJar")
                     cookies = cookies.get_dict()
@@ -162,6 +262,7 @@ class AlexaLogin:
                     for key, value in cookies.items():
                         if self._debug:
                             _LOGGER.debug('Key: "%s", Value: "%s"', key, value)
+                        # escape extra quote marks from Requests cookie
                         self._cookies[str(key)] = value.strip('"')
                     numcookies = len(self._cookies)
                 elif isinstance(cookies, defaultdict):
@@ -184,6 +285,18 @@ class AlexaLogin:
                     _LOGGER.debug("Found dict cookie")
                     self._cookies = cookies
                     numcookies = len(self._cookies)
+                elif isinstance(cookies, http.cookiejar.MozillaCookieJar):
+                    _LOGGER.debug("Found Mozillacookiejar")
+                    for cookie in cookies:
+                        if self._debug:
+                            _LOGGER.debug(
+                                "Processing cookie %s expires: %s",
+                                cookie,
+                                cookie.expires,
+                            )
+                        # escape extra quote marks from MozillaCookieJar cookie
+                        self._cookies[cookie.name] = cookie.value.strip('"')
+                    numcookies = len(self._cookies)
                 else:
                     _LOGGER.debug("Ignoring unknown file %s", type(cookies))
                 if numcookies:
@@ -202,7 +315,7 @@ class AlexaLogin:
                                 self._cookiefile[0],
                                 EXCEPTION_TEMPLATE.format(type(ex).__name__, ex.args),
                             )
-        await self.login(cookies=self._cookies)
+        return self._cookies
 
     async def close(self) -> None:
         """Close connection for login."""
@@ -260,8 +373,6 @@ class AlexaLogin:
         Returns false if no csrf found; necessary to issue commands
         """
         if self._debug:
-            from json import dumps
-
             _LOGGER.debug("Testing whether logged in to alexa.%s", self._url)
             _LOGGER.debug("Cookies: %s", cookies)
             _LOGGER.debug("Session Cookies:\n%s", self._print_session_cookies())
@@ -284,11 +395,15 @@ class AlexaLogin:
         self.customer_id = json.get("authentication", {}).get("customerId")
         if email != "" and email.lower() == self._email.lower():
             _LOGGER.debug("Logged in as %s with id: %s", email, self.customer_id)
+            self.stats["login_timestamp"] = datetime.datetime.now()
+            self.stats["api_calls"] = 0
             return True
         if email == "":
             _LOGGER.debug(
                 "Logged in as mobile account %s with %s", email, self.customer_id
             )
+            self.stats["login_timestamp"] = datetime.datetime.now()
+            self.stats["api_calls"] = 0
             return True
         _LOGGER.debug("Not logged in due to email mismatch")
         await self.reset()
@@ -298,16 +413,15 @@ class AlexaLogin:
         if not self._session or force:
             #  define session headers
             self._headers = {
-                "User-Agent": (
-                    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_14_6) "
-                    "AppleWebKit/537.36 (KHTML, like Gecko) "
-                    "Chrome/84.0.4147.135 Safari/537.36"
-                ),
-                "Accept": (
-                    "text/html,application/xhtml+xml, "
-                    "application/xml;q=0.9,*/*;q=0.8"
-                ),
+                "User-Agent": USER_AGENT,
+                # "User-Agent": (
+                #     "Mozilla/5.0 (iPhone; CPU iPhone OS 13_5_1 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Mobile/15E148 PitanguiBridge/2.2.345247.0-[HARDWARE=iPhone10_4][SOFTWARE=13.5.1]"
+                # ),
+                "Accept": ("*/*"),
                 "Accept-Language": "*",
+                "DNT": "1",
+                "Upgrade-Insecure-Requests": "1",
+                "authority": f"www.{self._url}",
             }
 
             #  initiate session
@@ -330,9 +444,9 @@ class AlexaLogin:
             )
         for cookie in cookie_jar:
             oldvalue = self._cookies[cookie.key] if cookie.key in self._cookies else ""
-            if cookie["domain"] == str(site):
+            if cookie["domain"] == str(site) or cookie["domain"] == "":
                 self._cookies[cookie.key] = cookie.value
-                if self._debug:
+                if self._debug and oldvalue != cookie.value:
                     _LOGGER.debug(
                         "%s: key: %s value: %s -> %s",
                         site,
@@ -346,13 +460,7 @@ class AlexaLogin:
         if not self._session.cookie_jar:
             result = "Session cookie jar is empty."
         for cookie in self._session.cookie_jar:
-            result += "{}: expires:{} max-age:{} {}={}\n".format(
-                cookie["domain"],
-                cookie["expires"],
-                cookie["max-age"],
-                cookie.key,
-                cookie.value,
-            )
+            result += f"{cookie}\n"
         return result
 
     @_catch_all_exceptions
@@ -371,14 +479,33 @@ class AlexaLogin:
                 self.status = {}
                 self.status["login_successful"] = True
                 _LOGGER.debug("Log in successful with cookies")
-                self._prepare_cookies_from_session(self._url)
+                await self.get_tokens()
+                await self.save_cookiefile()
                 return
             await self.reset()
-        _LOGGER.debug("No valid cookies for log in; using credentials")
-        #  site = 'https://www.' + self._url + '/gp/sign-in.html'
-        #  use alexa site instead
+        _LOGGER.debug("Using credentials to log in")
         if not self._site:
-            site: Text = self._prefix + self._url
+            site: URL = URL(f"https://www.{self._url}/ap/signin")
+            deviceid: Text = "453535333833343639344134304445313832303944354342344141414342453123413249564c5635564d32573831"
+            query = {
+                "openid.return_to": f"https://www.{self._url}/ap/maplanding",
+                "openid.assoc_handle": "amzn_dp_project_dee_ios",
+                "openid.identity": "http://specs.openid.net/auth/2.0/identifier_select",
+                "pageId": "amzn_dp_project_dee_ios",
+                "accountStatusPolicy": "P1",
+                "openid.claimed_id": "http://specs.openid.net/auth/2.0/identifier_select",
+                "openid.mode": "checkid_setup",
+                "openid.ns.oa2": f"http://www.{self._url}/ap/ext/oauth/2",
+                "openid.oa2.client_id": f"device:{deviceid}",
+                "openid.ns.pape": "http://specs.openid.net/extensions/pape/1.0",
+                "openid.oa2.response_type": "token",
+                "openid.ns": "http://specs.openid.net/auth/2.0",
+                "openid.pape.max_auth_age": "0",
+                "openid.oa2.scope": "device_auth_access",
+                "language": "en_US",
+            }
+            site = site.update_query(query)
+            _LOGGER.debug("Attempting oauth login to %s", site)
         else:
             site = self._site
         if not self._session:
@@ -416,19 +543,16 @@ class AlexaLogin:
             async with aiofiles.open(self._debugget, mode="wb") as localfile:
                 await localfile.write(await resp.read())
         # This commented block can be used to read a file directly to process.
-        # import aiofiles
-
-        # async with aiofiles.open("/config/anti-automation-js.html", "rb") as myfile:
         # async with aiofiles.open(
+        #     "/config/anti-automation-js.html", "rb"
         #     "/config/Amazon-Password-Assistance.html", "rb"
+        #     "/config/password_reset_required.html", "rb"
         # ) as myfile:
         #     html = await myfile.read()
         site = await self._process_page(html, site)
         if not self.status.get("ap_error"):
             missing_params = self._populate_data(site, data)
             if self._debug:
-                from json import dumps  # pylint: disable=import-outside-toplevel
-
                 if missing_params:
                     _LOGGER.debug(
                         "WARNING: Detected missing params: %s",
@@ -456,6 +580,241 @@ class AlexaLogin:
             site = await self._process_resp(post_resp)
             self._site = await self._process_page(await post_resp.text(), site)
 
+    async def save_cookiefile(self) -> None:
+        """Save login session cookies to file."""
+        for cookiefile in self._cookiefile:
+            if cookiefile == self._cookiefile[0]:
+                cookie_jar = self._session.cookie_jar
+                assert isinstance(cookie_jar, aiohttp.CookieJar)
+                cookie_jar.update_cookies(self._cookies, URL(self._url))
+                self._prepare_cookies_from_session(self._url)
+                if self._debug:
+                    _LOGGER.debug("Saving cookie to %s", cookiefile)
+                try:
+                    cookie_jar.save(self._cookiefile[0])
+                except (OSError, EOFError, TypeError, AttributeError) as ex:
+                    _LOGGER.debug(
+                        "Error saving pickled cookie to %s: %s",
+                        self._cookiefile[0],
+                        EXCEPTION_TEMPLATE.format(type(ex).__name__, ex.args),
+                    )
+            elif (cookiefile) and os.path.exists(cookiefile):
+                _LOGGER.debug("Removing outdated cookiefile %s", cookiefile)
+                await delete_cookie(cookiefile)
+
+    async def get_tokens(self) -> bool:
+        """Get access and refresh tokens after registering device using cookies.
+
+        Returns
+            bool: True if successful.
+
+        """
+
+        if not self.access_token:
+            _LOGGER.warning(
+                "No access token found; falling back to credential login instead of oauth."
+            )
+            return False
+        frc = base64.b64encode(secrets.token_bytes(313)).decode("ascii")
+        map_md = base64.b64encode(
+            '{"device_user_dictionary":[],"device_registration_data":{"software_version":"1"},"app_identifier":{"app_version":"2.2.223830","bundle_id":"com.amazon.echo"}}'.encode(
+                "utf8"
+            )
+        ).decode("utf8")
+        self._cookies["frc"] = frc
+        self._cookies["map-md"] = map_md
+        headers = {
+            "Content-Type": "application/json",
+            "Accept-Charset": "utf-8",
+            "x-amzn-identity-auth-domain": f"api.{self._url}",
+            "Connection": "keep-alive",
+            "Accept": "*/*",
+            "User-Agent": USER_AGENT,
+            "Accept-Language": "en-US",
+            # "Cookie": "; ".join(
+            #     [str(x) + "=" + str(y) for x, y in self._cookies.items()]
+            # ),
+        }
+        cookies = []
+        for k, value in self._cookies.items():
+            # if k == "csrf":
+            #     continue
+            cookies.append({"Value": value, "Name": k})
+        data = {
+            "requested_extensions": ["device_info", "customer_info"],
+            "cookies": {"website_cookies": cookies, "domain": f".{self._url}"},
+            "registration_data": {
+                "domain": "Device",
+                "app_version": "2.2.223830.0",
+                "device_type": "A2IVLV5VM2W81",
+                "device_name": f"%FIRST_NAME%'s%DUPE_STRATEGY_1ST%{APP_NAME}",
+                "os_version": "11.4.1",
+                "device_serial": "2cf947a5a093d686a30c33cb07a703fd",
+                # may need to be replaced with random 16 bytes; leaving static for now
+                # https://github.com/Apollon77/alexa-cookie/blob/master/lib/proxy.js#L102-L106
+                "device_model": "iPhone",
+                "app_name": APP_NAME,
+                "software_version": "1",
+            },
+            "auth_data": {"access_token": self.access_token},
+            "user_context_map": {"frc": frc},
+            "requested_token_type": ["bearer", "mac_dms", "website_cookies"],
+        }
+
+        response = await self._session.post(
+            "https://api." + self._url + "/auth/register", json=data, headers=headers,
+        )
+        if response.status != 200:
+            return False
+        response = (await response.json()).get("response")
+        # _LOGGER.debug("auth response %s with \n%s", response, dumps(data))
+        if response.get("success"):
+            _LOGGER.debug("Successfully registered device with Amazon")
+            if self._debug:
+                _LOGGER.debug("Received registration data:\n%s", dumps(response))
+            self.refresh_token = response["success"]["tokens"]["bearer"][
+                "refresh_token"
+            ]
+            old = self.access_token
+            self.access_token = response["success"]["tokens"]["bearer"]["access_token"]
+            self.expires_in = datetime.datetime.now().timestamp() + int(
+                response["success"]["tokens"]["bearer"]["expires_in"]
+            )
+            if old != self.access_token:
+                _LOGGER.debug(
+                    "New access token received which expires at %s in %s",
+                    datetime.datetime.fromtimestamp(self.expires_in),
+                    datetime.datetime.fromtimestamp(self.expires_in)
+                    - datetime.datetime.now(),
+                )
+            return True
+        return False
+
+    async def refresh_access_token(self) -> bool:
+        """Refresh access token and expires in using refresh token.
+
+        Returns
+            bool: Return true if successful.
+
+        """
+        if not self.refresh_token:
+            _LOGGER.debug("No refresh token found to get access_token")
+            return False
+        data = {
+            "app_name": APP_NAME,
+            "app_version": "2.2.223830.0",
+            "di.sdk.version": "6.10.0",
+            "source_token": self.refresh_token,
+            "package_name": "com.amazon.echo",
+            "di.hw.version": "iPhone",
+            "platform": "iOS",
+            "requested_token_type": "access_token",
+            "source_token_type": "refresh_token",
+            "di.os.name": "iOS",
+            "di.os.version": "11.4.1",
+            "current_version": "6.10.0",
+        }
+        headers = {
+            "Content-Type": "application/x-www-form-urlencoded",
+            "Accept-Charset": "utf-8",
+            "x-amzn-identity-auth-domain": f"api.{self._url}",
+            "Connection": "keep-alive",
+            "Accept": "*/*",
+            "User-Agent": USER_AGENT,
+            "Accept-Language": "en-US",
+            "Cookie": "; ".join(
+                [str(x) + "=" + str(y) for x, y in self._cookies.items()]
+            ),
+        }
+        response = await self._session.post(
+            "https://api." + self._url + "/auth/token", data=data, headers=headers,
+        )
+        # _LOGGER.debug("refresh token response %s with \n%s", response, dumps(data))
+        if response.status != 200:
+            return False
+        response = await response.json()
+        # _LOGGER.debug("refresh token json %s ", response)
+        if response.get("access_token"):
+            self.access_token = response.get("access_token")
+            self.expires_in = datetime.datetime.now().timestamp() + int(
+                response.get("expires_in")
+            )
+            _LOGGER.debug(
+                "Successfully updated access_token which expires at %s in %s",
+                datetime.datetime.fromtimestamp(self.expires_in),
+                datetime.datetime.fromtimestamp(self.expires_in)
+                - datetime.datetime.now(),
+            )
+
+    async def exchange_token_for_cookies(self) -> bool:
+        """Generate new session cookies using refresh token.
+
+        Returns
+            bool: True if succesful
+
+        """
+        if not self.refresh_token:
+            _LOGGER.debug("No refresh token found to get access token")
+            return False
+        data = {
+            "di.os.name": "iOS",
+            "app_version": "2.2.223830.0",
+            "domain": f".{self.url}",
+            "source_token": self.refresh_token,
+            "requested_token_type": "auth_cookies",
+            "source_token_type": "refresh_token",
+            "di.hw.version": "iPhone",
+            "di.sdk.version": "6.10.0",
+            "cookies": {},
+            "app_name": APP_NAME,
+            "di.os.version": "11.4.1",
+        }
+        headers = {
+            "Content-Type": "application/x-www-form-urlencoded",
+            "Accept-Charset": "utf-8",
+            "x-amzn-identity-auth-domain": f"api.{self._url}",
+            "Connection": "keep-alive",
+            "Accept": "*/*",
+            "User-Agent": USER_AGENT,
+            "Accept-Language": "en-US",
+            # "Cookie": "; ".join(
+            #     [str(x) + "=" + str(y) for x, y in self._cookies.items()]
+            # ),
+        }
+        response = await self._session.post(
+            "https://api." + self._url + "/ap/exchangetoken/cookies",
+            data=data,
+            headers=headers,
+        )
+        # _LOGGER.debug("exchange token response %s with \n%s", response, dumps(data))
+        if response.status != 200:
+            return False
+        response = (await response.json()).get("response")
+        # _LOGGER.debug("exchange token json %s ", response)
+        for domain, cookies in response["tokens"]["cookies"].items():
+            # _LOGGER.debug("updating %s with %s", domain, cookies)
+            for item in cookies:
+                raw_cookie = SimpleCookie()
+                cookie_name = item["Name"]
+                raw_cookie[item["Name"]] = (
+                    item["Value"]
+                    if not (
+                        item["Value"].startswith('"') and item["Value"].endswith('"')
+                    )
+                    # Strings are returned within quotations, strip them
+                    else item["Value"][1:-1]
+                )
+                for name, value in item.items():
+                    if name in ["Name", "Value"]:
+                        continue
+                    raw_cookie[cookie_name][name] = value
+                # _LOGGER.debug("updating jar with cookie %s", raw_cookie)
+                self._session.cookie_jar.update_cookies(raw_cookie, URL(domain))
+        _LOGGER.debug(
+            "Cookies successfully exchanged for refresh token for domain %s",
+            response["tokens"]["cookies"].keys(),
+        )
+
     async def _process_resp(self, resp) -> Text:
         if resp.history:
             for item in resp.history:
@@ -538,6 +897,7 @@ class AlexaLogin:
             self._links = links
 
         _LOGGER.debug("Processing %s", site)
+        site_url = URL(site)
         soup: BeautifulSoup = BeautifulSoup(html, "html.parser")
 
         status: Dict[Text, Union[Text, bool]] = {}
@@ -658,7 +1018,7 @@ class AlexaLogin:
             _LOGGER.debug("Verification code requested:")
             status["verificationcode_required"] = True
             self._data = self.get_inputs(soup, {"action": "verify"})
-        elif missingcookies_tag is not None:
+        elif missingcookies_tag is not None and site_url.path != "/ap/maplanding":
             _LOGGER.debug("Error page detected:")
             href = ""
             links = missingcookies_tag.findAll("a", href=True)
@@ -676,7 +1036,7 @@ class AlexaLogin:
             status["force_get"] = True
             status["message"] = re.sub("(\\s)+", "\\1", message)
             _LOGGER.debug("Javascript Authentication page detected: %s", message)
-        elif forgotpassword_tag:
+        elif forgotpassword_tag or soup.find("input", {"name": "OTPChallengeOptions"}):
             status["message"] = (
                 "Forgot password page detected; "
                 "Amazon has detected too many failed logins. "
@@ -697,36 +1057,21 @@ class AlexaLogin:
         else:
             _LOGGER.debug("Captcha/2FA not requested; confirming login.")
             if await self.test_loggedin():
+                query = site_url.query
+                self.access_token = query.get("openid.oa2.access_token")
                 _LOGGER.debug(
                     "Login confirmed; saving cookie to %s", self._cookiefile[0]
                 )
                 status["login_successful"] = True
-                self._prepare_cookies_from_session(self._url)
-                if self._debug:
-                    _LOGGER.debug("Saving cookie: %s", self._print_session_cookies())
-                for cookiefile in self._cookiefile:
-                    if cookiefile == self._cookiefile[0]:
-                        cookie_jar = self._session.cookie_jar
-                        assert isinstance(cookie_jar, aiohttp.CookieJar)
-                        try:
-                            cookie_jar.save(self._cookiefile[0])
-                        except (OSError, EOFError, TypeError, AttributeError) as ex:
-                            _LOGGER.debug(
-                                "Error saving pickled cookie to %s: %s",
-                                self._cookiefile[0],
-                                EXCEPTION_TEMPLATE.format(type(ex).__name__, ex.args),
-                            )
-                    elif (cookiefile) and os.path.exists(cookiefile):
-                        _LOGGER.debug("Removing outdated cookiefile %s", cookiefile)
-                        await delete_cookie(cookiefile)
+                await self.get_tokens()
+                await self.save_cookiefile()
                 #  remove extraneous Content-Type to avoid 500 errors
                 self._headers.pop("Content-Type", None)
 
             else:
                 _LOGGER.debug("Login failed; check credentials")
                 status["login_failed"] = "login_failed"
-                assert self._data is not None
-                if "" in self._data.values():
+                if self._data and "" in self._data.values():
                     missing = [k for (k, v) in self._data.items() if v == ""]
                     _LOGGER.debug(
                         "If credentials correct, please report"
@@ -770,16 +1115,21 @@ class AlexaLogin:
 
     def _populate_data(self, site: Text, data: Dict[str, Optional[str]]) -> bool:
         """Populate self._data with info from data."""
-        # pull data from configurator
-        password: Optional[Text] = data.get("password")
-        captcha: Optional[Text] = data.get("captcha")
-        securitycode: Optional[Text] = data.get("securitycode")
-        claimsoption: Optional[Text] = data.get("claimsoption")
-        authopt: Optional[Text] = data.get("authselectoption")
-        verificationcode: Optional[Text] = data.get("verificationcode")
         _LOGGER.debug(
             "Preparing form submission to %s with input data: %s", site, obfuscate(data)
         )
+        # pull data from configurator
+        password: Optional[Text] = data.get("password")
+        captcha: Optional[Text] = data.get("captcha")
+        if data.get("otp_secret"):
+            self.set_totp(data.get("otp_secret"))
+        securitycode: Optional[Text] = data.get("securitycode")
+        if not securitycode and self._totp:
+            _LOGGER.debug("No 2FA code supplied but will generate.")
+            securitycode = self.get_totp_token()
+        claimsoption: Optional[Text] = data.get("claimsoption")
+        authopt: Optional[Text] = data.get("authselectoption")
+        verificationcode: Optional[Text] = data.get("verificationcode")
 
         #  add username and password to self._data for post request
         #  self._data is scraped from the form page in _process_page
@@ -787,13 +1137,14 @@ class AlexaLogin:
         if self._data:
             if "email" in self._data and self._data["email"] == "":
                 self._data["email"] = self._email
-            if "password" in self._data and self._data["password"] == "":
                 # add the otp to the password if available
-                self._data["password"] = (
-                    self._password
-                    if not password
-                    else password + data.get("securitycode", "")
-                )
+            self._data["password"] = (
+                self._password + securitycode
+                if not password
+                else password + securitycode
+                if securitycode
+                else password
+            )
             if "rememberMe" in self._data:
                 self._data["rememberMe"] = "true"
             if captcha is not None and "guess" in self._data:
