@@ -14,12 +14,13 @@ from http.cookies import CookieError
 from json import JSONDecodeError
 import logging
 import os
+from types import MappingProxyType
+from typing import Optional, Text, Union
 
 import aiofiles.os as aioos
 
-from typing import Optional, Union, Text
-
 from alexapy.aiohttp import ClientConnectionError, ContentTypeError
+import alexapy.alexalogin
 
 from .const import EXCEPTION_TEMPLATE
 from .errors import (
@@ -53,15 +54,24 @@ def hide_serial(item: Optional[Union[dict, str, list]]) -> Union[dict, str, list
     if isinstance(item, dict):
         response = item.copy()
         for key, value in item.items():
-            if isinstance(value, (dict, list)) or key in [
-                "deviceSerialNumber",
-                "serialNumber",
-                "destinationUserId",
-                "customerId",
-            ]:
+            if (
+                isinstance(value, (dict, list))
+                or key
+                in [
+                    "deviceSerialNumber",
+                    "serialNumber",
+                    "destinationUserId",
+                    "customerId",
+                ]
+                or "secret" in key
+            ):
                 response[key] = hide_serial(value)
     elif isinstance(item, str):
-        response = "{}{}{}".format(item[0], "*" * (len(item) - 4), item[-3:])
+        response = (
+            "{}{}{}".format(item[0], "*" * (len(item) - 4), item[-3:])
+            if len(item) > 6
+            else f"{'*' * len(item)}"
+        )
     elif isinstance(item, list):
         response = []
         for list_item in item:
@@ -76,19 +86,25 @@ def obfuscate(item):
     """Obfuscate email, password, and other known sensitive keys."""
     if item is None:
         return ""
-    if isinstance(item, dict):
+    if isinstance(item, (MappingProxyType, dict)):
         response = item.copy()
         for key, value in item.items():
             if key in ["password"]:
                 response[key] = f"REDACTED {len(value)} CHARS"
             elif key in ["email"]:
                 response[key] = hide_email(value)
-            elif key in [
-                "deviceSerialNumber",
-                "serialNumber",
-                "destinationUserId",
-                "customerId",
-            ]:
+            elif key in ["cookies_txt"]:
+                response[key] = "OBFUSCATED COOKIE"
+            elif (
+                key
+                in [
+                    "deviceSerialNumber",
+                    "serialNumber",
+                    "destinationUserId",
+                    "customerId",
+                ]
+                or "secret" in key
+            ):
                 response[key] = hide_serial(value)
             elif isinstance(value, (dict, list, tuple)):
                 response[key] = obfuscate(value)
@@ -112,6 +128,11 @@ def _catch_all_exceptions(func):
 
     @functools.wraps(func)
     async def wrapper(*args, **kwargs):
+        login: Optional["alexapy.alexalogin.AlexaLogin"] = None
+        for arg in args:
+            if isinstance(arg, alexapy.alexalogin.AlexaLogin):
+                login = arg
+                break
         try:
             return await func(*args, **kwargs)
         except (ClientConnectionError, KeyError) as ex:
@@ -123,7 +144,7 @@ def _catch_all_exceptions(func):
                 obfuscate(kwargs),
                 EXCEPTION_TEMPLATE.format(type(ex).__name__, ex.args),
             )
-            raise AlexapyConnectionError
+            raise AlexapyConnectionError from ex
         except (JSONDecodeError, CookieError) as ex:
             _LOGGER.warning(
                 "%s.%s(%s, %s): A login error occured: %s",
@@ -133,7 +154,8 @@ def _catch_all_exceptions(func):
                 obfuscate(kwargs),
                 EXCEPTION_TEMPLATE.format(type(ex).__name__, ex.args),
             )
-            raise AlexapyLoginError
+            login.status["login_successful"] = False
+            raise AlexapyLoginError from ex
         except (ContentTypeError) as ex:
             _LOGGER.warning(
                 "%s.%s(%s, %s): A login error occured; Amazon may want you to change your password: %s",
@@ -143,7 +165,8 @@ def _catch_all_exceptions(func):
                 obfuscate(kwargs),
                 EXCEPTION_TEMPLATE.format(type(ex).__name__, ex.args),
             )
-            raise AlexapyLoginError
+            login.status["login_successful"] = False
+            raise AlexapyLoginError from ex
         except CancelledError as ex:
             _LOGGER.warning(
                 "%s.%s(%s, %s): Timeout error occured accessing AlexaAPI: %s",
